# Copyright 2000-2020 JetBrains s.r.o.
# Use of this source code is governed by the Apache 2.0 license that can be found
# in the LICENSE file.


"""Run configurations related functions"""

from os import listdir, mkdir, path, rename
from os.path import join, isdir
from shutil import rmtree
from typing import Optional, Dict, List
import configparser

from .apps import get_app_path, make_run_script
from .global_config import get_run_configs_dir, RunConfig
from .ide_configuration import install_own_markdown_plugin
from .secure_config import generate_server_secrets, is_secure

CONFIG_INI_NAME = 'config.ini'
RUN_SCRIPT_NAME = 'run.sh'


def load_config(config_name: str) -> RunConfig:
    """Loads specified config from disk."""
    config = configparser.ConfigParser()
    config_path = join(get_run_configs_dir(), config_name, CONFIG_INI_NAME)
    config.read(config_path)

    return RunConfig(config_name,
                     config.get('IDE', 'PATH'),
                     config.getint('PROJECTOR', 'PORT'),
                     config.get('HTTP.SERVER', 'ADDRESS'),
                     config.getint('HTTP.SERVER', 'PORT'),
                     config.get('SSL', 'TOKEN', fallback=''))


def get_run_script_path(config_name: str) -> str:
    """Returns full path to projector run script"""
    return join(get_run_configs_dir(), config_name, RUN_SCRIPT_NAME)


def generate_run_script(run_config: RunConfig) -> None:
    """Generates projector run script"""
    run_script = get_run_script_path(run_config.name)
    make_run_script(run_config, run_script)


def save_config(run_config: RunConfig) -> None:
    """Saves given run config."""
    config = configparser.ConfigParser()
    config['IDE'] = {}
    config['IDE']['PATH'] = run_config.path_to_app

    config['PROJECTOR'] = {}
    config['PROJECTOR']['PORT'] = str(run_config.projector_port)

    config['HTTP.SERVER'] = {}
    config['HTTP.SERVER']['ADDRESS'] = run_config.http_address
    config['HTTP.SERVER']['PORT'] = str(run_config.http_port)

    if is_secure(run_config):
        config['SSL'] = {}
        config['SSL']['TOKEN'] = run_config.token

    config_path = join(get_run_configs_dir(), run_config.name)

    if not path.isdir(config_path):
        mkdir(config_path)

    config_path = join(config_path, CONFIG_INI_NAME)

    with open(config_path, 'w') as configfile:
        config.write(configfile)

    generate_run_script(run_config)

    if is_secure(run_config):
        generate_server_secrets(run_config)


def get_run_configs(pattern: Optional[str] = None) -> Dict[str, RunConfig]:
    """Get run configs, matched given pattern."""
    res = {}

    for config_name in listdir(get_run_configs_dir()):
        if pattern and config_name.lower().find(pattern.lower()) == -1:
            continue

        config = load_config(config_name)

        if pattern == config_name:
            return {config_name: config}

        res[config_name] = config

    return res


def get_run_config_names(pattern: Optional[str] = None) -> List[str]:
    """Get sorted run config names list, matched to given pattern."""
    res = list(get_run_configs(pattern).keys())
    res.sort()
    return res


def delete_config(config_name: str) -> None:
    """Removes specified config."""
    config_path = join(get_run_configs_dir(), config_name)
    rmtree(config_path, ignore_errors=True)


def rename_config(from_name: str, to_name: str) -> None:
    """Renames config from_name to to_name."""
    from_path = join(get_run_configs_dir(), from_name)
    to_path = join(get_run_configs_dir(), to_name)
    rename(from_path, to_path)


def make_config_name(app_name: str) -> str:
    """Creates config name from application name."""
    pos = app_name.find(' ')

    if pos != -1:
        return app_name[0:pos]

    return app_name


def validate_run_config(run_config: RunConfig) -> None:
    """Checks given config for validity."""
    if not isdir(run_config.path_to_app):
        raise ValueError(f'IDE path does not exist: {run_config.path_to_app}')

    if run_config.http_port == run_config.projector_port:
        raise ValueError('HTTP port can\'t be equal to projector port.')


def get_used_http_ports() -> List[int]:
    """Returns list of ports, used by http servers in existing configs."""
    return [rc.http_port for rc in get_run_configs().values()]


def get_used_projector_ports() -> List[int]:
    """Returns list of ports, used by projector servers in existing configs."""
    return [rc.projector_port for rc in get_run_configs().values()]


def get_used_ports() -> List[int]:
    """Returns list of ports, used either by projector or http servers in existing configs."""
    return get_used_http_ports() + get_used_projector_ports()


def get_configs_with_app(app_name: str) -> List[str]:
    """Returns list of configs which referees to given app name."""
    app_path = get_app_path(app_name)
    return [k for k, v in get_run_configs().items() if v.path_to_app == app_path]


def update_markdown_plugin(run_config: RunConfig) -> None:
    """
    Updates markdown plugin in specified config with bundled version.
    Useful after script update.
    """
    install_own_markdown_plugin(run_config.path_to_app)



# Copyright 2000-2020 JetBrains s.r.o.
# Use of this source code is governed by the Apache 2.0 license that can be found
# in the LICENSE file.


"""Application management functions."""

from os.path import join, expanduser, dirname
from os import listdir, chmod, stat
from typing import Optional, List
import json

from .global_config import get_apps_dir, get_projector_server_dir, COMPATIBLE_APPS, \
    CompatibleApp, RunConfig
from .secure_config import is_secure, get_ssl_properties_file, SSL_ENV_NAME
from .utils import unpack_tar_file

IDEA_RUN_CLASS = 'com.intellij.idea.Main'
PROJECTOR_RUN_CLASS = 'org.jetbrains.projector.server.ProjectorLauncher'
IDEA_PLATFORM_PREFIX = 'idea.platform.prefix'
IDEA_PATH_SELECTOR = 'idea.paths.selector'


def unpack_app(file_path: str) -> str:
    """Unpacks specified file to app directory."""
    return unpack_tar_file(file_path, get_apps_dir())


def get_installed_apps(pattern: Optional[str] = None) -> List[str]:
    """Returns sorted list of installed apps, matched given pattern."""
    res = [file_name for file_name in listdir(get_apps_dir()) if
           pattern is None or file_name.lower().find(pattern.lower()) != -1]
    res.sort()
    return res


def get_compatible_apps(pattern: Optional[str] = None) -> List[CompatibleApp]:
    """Returns list of compatible apps, matched given pattern."""
    apps = [app for app in COMPATIBLE_APPS if
            pattern is None or app.name.lower().find(pattern.lower()) != -1]

    if pattern:
        for app in apps:
            if pattern.lower() == app.name.lower():
                return [app]

    return apps


def get_compatible_app_names(pattern: Optional[str] = None) -> List[str]:
    """Get sorted list of projector-compatible applications, matches given pattern."""
    res = [app.name for app in get_compatible_apps(pattern)]
    res.sort()
    return res


def get_app_path(app_name: str) -> str:
    """Returns full path to given app."""
    return join(get_apps_dir(), app_name)


def make_run_script(run_config: RunConfig, run_script: str) -> None:
    """Creates run script from ide launch script."""
    idea_script = get_launch_script(run_config.path_to_app)

    with open(idea_script, 'r') as src, open(run_script, 'w') as dst:
        for line in src:
            if line.startswith("IDE_BIN_HOME"):
                line = f'IDE_BIN_HOME={join(run_config.path_to_app, "bin")}\n'
            elif line.find("classpath") != -1:
                line = f' -classpath "$CLASSPATH:{get_projector_server_dir()}/*" \\\n'
            elif line.find(IDEA_RUN_CLASS) != -1:
                line = f' -Dorg.jetbrains.projector.server.port={run_config.projector_port} \\\n'
                line += f' -Dorg.jetbrains.projector.server.classToLaunch={IDEA_RUN_CLASS} \\\n'

                if is_secure(run_config):
                    line += f' -D{SSL_ENV_NAME}=\"{get_ssl_properties_file(run_config.name)}\" \\\n'
                    # line += f' -D{TOKEN_ENV_NAME}=\"{run_config.token}\" \\\n'

                line += f'  {PROJECTOR_RUN_CLASS}\\\n'

            dst.write(line)

    stats = stat(run_script)
    chmod(run_script, stats.st_mode | 0o0111)


class ProductInfo:
    """Ide product info"""

    # pylint: disable=too-many-instance-attributes
    def __init__(self, name: str, version: str, build_number: str, product_code: str,
                 data_dir: str, svg_icon_path: str, os: str, launcher_path: str,
                 java_exec_path: str, vm_options_path: str, startup_wm_class: str) -> None:
        self.name: str = name
        self.version: str = version
        self.build_number: str = build_number
        self.product_code: str = product_code
        self.data_dir: str = data_dir
        self.svg_icon_path: str = svg_icon_path
        self.os: str = os
        self.launcher_path: str = launcher_path
        self.java_exec_path: str = java_exec_path
        self.vm_options_path: str = vm_options_path
        self.startup_wm_class: str = startup_wm_class


PRODUCT_INFO = 'product-info.json'


class Version:
    """Three-parts version representation."""

    def __init__(self, year: int, quart: int, last: int):
        self.year: int = year
        self.quart: int = quart
        self.last: int = last


def parse_version(version: str) -> Version:
    """Parses version string to Version class."""
    parsed = version.split(".")
    return Version(int(parsed[0]), int(parsed[1]), int(parsed[2] if len(parsed) > 2 else -1))


def get_data_dir_from_script(run_script: str) -> str:
    """Returns idea data dir from run script."""
    with open(run_script, 'r') as file:
        for line in file:
            pos = line.find(IDEA_PATH_SELECTOR)

            if pos != -1:
                parts = line.split('=')

                if len(parts) < 2:
                    raise Exception(f'Unable to parse {IDEA_PATH_SELECTOR} line.')

                return parts[1].split(' ')[0]

    raise Exception('Unable to find data directory in the launch script.')


def get_product_info(app_path: str) -> ProductInfo:
    """Parses product info file to ProductInfo class."""
    prod_info_path = join(app_path, PRODUCT_INFO)
    with open(prod_info_path, "r") as file:
        data = json.load(file)

        product_info = ProductInfo(data['name'], data['version'], data['buildNumber'],
                                   data['productCode'], '', data['svgIconPath'],
                                   data['launch'][0]['os'],
                                   data['launch'][0]['launcherPath'],
                                   data['launch'][0]['javaExecutablePath'],
                                   data['launch'][0]['vmOptionsFilePath'],
                                   data['launch'][0]['startupWmClass'])

        version = parse_version(product_info.version)

        if version.year >= 2020 and version.quart >= 2:
            product_info.data_dir = data['dataDirectoryName']
        else:
            product_info.data_dir = get_data_dir_from_script(
                join(app_path, product_info.launcher_path))

        return product_info


def get_launch_script(app_path: str) -> str:
    """Returns full path to launch script by ide path."""
    prod_info = get_product_info(app_path)
    return join(app_path, prod_info.launcher_path)


def get_bin_dir(app_path: str) -> str:
    """Get full path to ide bin dir."""
    run_script = get_launch_script(app_path)
    return dirname(run_script)


CONFIG_PREFIX = expanduser('~/')
VER_2020_CONFIG_PREFIX = expanduser('~/.config/JetBrains')


def get_config_dir(app_path: str) -> str:
    """Returns ide config directory."""
    product_info = get_product_info(app_path)
    version = parse_version(product_info.version)

    if version.year >= 2020:
        return join(VER_2020_CONFIG_PREFIX, product_info.data_dir)

    return join(join(CONFIG_PREFIX, '.' + product_info.data_dir), 'config')


PLUGIN_2020_PREFIX: str = expanduser('~/.local/share/JetBrains')


def get_plugin_dir(app_path: str) -> str:
    """Returns full path to application plugin directory."""
    product_info = get_product_info(app_path)
    version = parse_version(product_info.version)

    if version.year >= 2020:
        return join(PLUGIN_2020_PREFIX, product_info.data_dir)

    return join(get_config_dir(app_path), "plugins")



# Copyright 2000-2020 JetBrains s.r.o.
# Use of this source code is governed by the Apache 2.0 license that can be found
# in the LICENSE file.


"""User dialog related procedures."""

import sys
from os.path import expanduser
from typing import Optional, Dict, List, Tuple

import click
import netifaces  # type: ignore

from .apps import get_installed_apps, get_app_path, get_compatible_app_names
from .run_config import get_run_configs, RunConfig, get_run_config_names, \
    get_used_projector_ports, get_used_http_ports

from .global_config import DEF_HTTP_PORT, DEF_PROJECTOR_PORT
from .secure_config import generate_token


def display_run_configs_names(config_names: List[str]) -> None:
    """Pretty-print config names."""
    for i, config_name in enumerate(config_names):
        print(f'\t{i + 1:4}. {config_name}')


def list_configs(pattern: Optional[str] = None) -> None:
    """Displays names of run configs, matches to pattern."""
    config_names = get_run_config_names(pattern)
    display_run_configs_names(config_names)


def display_run_configs(run_configs: Dict[str, RunConfig]) -> None:
    """Display sorted names of given run configs."""
    config_names = list(run_configs.keys())
    config_names.sort()
    display_run_configs_names(config_names)


def find_apps(pattern: Optional[str] = None) -> None:
    """Pretty-print projector-compatible applications, matched to given pattern."""
    app_names = get_compatible_app_names(pattern)
    for i, app_name in enumerate(app_names):
        print(f'\t{i + 1:4}. {app_name}')


def list_apps(pattern: Optional[str] = None) -> None:
    """Pretty print the list of installed ide."""
    for i, app in enumerate(get_installed_apps(pattern)):
        print(f'\t{i + 1:4}. {app}')


def select_installed_app(pattern: Optional[str] = None) -> Optional[str]:
    """Interactively selects installed ide."""
    apps: List[str] = get_installed_apps(pattern)

    while True:
        list_apps(pattern)
        prompt = f'Choose an IDE number to uninstall or 0 to exit: [0-{len(apps)}]'
        app_number: int = click.prompt(prompt, type=int)

        if app_number < 0 or app_number > len(apps):
            print('Invalid number.')
            continue

        if app_number == 0:
            return None

        return apps[app_number - 1]


def select_compatible_app(pattern: Optional[str] = None) -> Optional[str]:
    """Interactively selects app name from list of projector-compatible applications."""
    app_names: List[str] = get_compatible_app_names(pattern)

    while True:
        find_apps(pattern)
        prompt = f'Choose IDE number to install or 0 to exit: [0-{len(app_names)}]'
        app_number: int = click.prompt(prompt, type=int)

        if app_number < 0 or app_number > len(app_names):
            print('Invalid number.')
            continue

        if app_number == 0:
            return None

        return app_names[app_number - 1]


def select_unused_config_name(hint: str) -> str:
    """Generate unused config name."""
    run_configs = get_run_configs()
    cnt = 0
    res = hint

    while res in run_configs:
        cnt += 1
        res = f'{hint}_{cnt}'

    return res


def select_new_config_name(hint: Optional[str]) -> Optional[str]:
    """Prompts for new config name and checks for collisions."""
    run_config = get_run_configs()

    if hint:
        hint = select_unused_config_name(hint)
        prompt = 'Enter a new configuration name or press ENTER for default'
    else:
        prompt = 'Enter a new configuration name'

    while True:
        name: str = click.prompt(prompt, default=hint)

        if not name:
            return None

        if name in run_config:
            print(f'A configuration with name {name} already exists, please choose another name.')
            print('The known configurations:')
            list_configs()
            continue

        return name


def select_run_config(config_name: Optional[str]) -> Tuple[str, RunConfig]:
    """Interactively select run config."""
    run_configs: Dict[str, RunConfig] = get_run_configs(config_name)

    if len(run_configs) == 0:
        print(f'Configuration with name {config_name} is unknown, exiting...')
        sys.exit(1)

    if len(run_configs) > 1:
        while True:
            display_run_configs(run_configs)
            prompt = f'Choose a configuration number or 0 to exit: [0-{len(run_configs)}]'
            config_number = click.prompt(prompt, type=int)

            if config_number < 0 or config_number > len(run_configs):
                print('Invalid number selected.')
                continue

            if config_number == 0:
                print('Configuration was not selected, exiting...')
                sys.exit(1)
            else:
                config_names = get_run_config_names(config_name)
                name = config_names[config_number - 1]
                return name, run_configs[name]

    return list(run_configs.items())[0]


def select_installed_app_path() -> Optional[str]:
    """Selects installed app and returns path to it."""
    apps = get_installed_apps()

    while True:
        list_apps()
        prompt = f'Choose IDE number to install or 0 to exit: [0-{len(apps)}]'
        app_number = click.prompt(prompt, type=int)

        if app_number < 0 or app_number > len(apps):
            print('Invalid number selected.')
            continue

        if app_number == 0:
            return None

        return get_app_path(apps[app_number - 1])


def select_manual_app_path() -> str:
    """Prompts for path to ide."""
    path: str = click.prompt('Enter the path to IDE', type=str)
    return path


def select_app_path() -> Optional[str]:
    """Select path to ide."""
    apps = get_installed_apps()

    if apps:
        inst = click.prompt('Do you want to choose a Projector-installed IDE? [y/n]', type=bool)

        if inst:
            return select_installed_app_path()

        return select_manual_app_path()

    enter = click.prompt('There are no installed Projector IDEs.\nWould you like to specify '
                         'a path to IDE manually?',
                         type=bool)
    if enter:
        return select_manual_app_path()

    return None


def get_all_listening_ports() -> List[int]:
    """
    Returns all tcp port numbers in LISTEN state (on any address).
    Reads port state from /proc/net/tcp.
    """
    res = []

    with open('/proc/net/tcp', 'r') as file:
        try:
            next(file)
            for line in file:
                split_line = line.strip().split(' ')
                hex_port = split_line[1].split(':')[1]
                hex_state = split_line[3]

                if hex_state == '0A':
                    res.append(int(hex_port, 16))
        except StopIteration:
            pass

    return res


def is_open_port(port: int) -> bool:
    """ Checks if tcp port is LISTEN state in system."""
    return port in get_all_listening_ports()


def get_def_port(ports: List[int], default: int) -> int:
    """Returns default or first unused in system and in run configs port."""
    port = default

    if len(ports) > 0:
        ports.sort()
        port = ports[-1] + 1

    while is_open_port(port):
        port += 1

    return port


def get_def_http_port() -> int:
    """Returns unused port for http server."""
    http_ports = get_used_http_ports()
    return get_def_port(http_ports, DEF_HTTP_PORT)


def get_def_projector_port() -> int:
    """Returns unused port for projector server."""
    ports = get_used_projector_ports()
    return get_def_port(ports, DEF_PROJECTOR_PORT)


def get_local_addresses() -> List[str]:
    """Returns list of local ip addresses."""
    interfaces = netifaces.interfaces()
    res = []

    for ifs in interfaces:
        addresses = netifaces.ifaddresses(ifs)

        if netifaces.AF_INET in addresses:
            ipv4 = addresses[netifaces.AF_INET]

            for ips in ipv4:
                res.append(ips['addr'])

    return res


def check_listening_address(address: str) -> bool:
    """Check entered ip address for validity."""
    if address == 'localhost':
        return True

    return address in get_local_addresses()


def select_http_address(default: str) -> str:
    """Selects address for http listening."""
    while True:
        res: str = click.prompt('Enter HTTP listening address (press ENTER for default)',
                                default=default)

        if check_listening_address(res):
            return res

        click.echo('You entered incorrect address, please try again.')
        click.echo('You can try to use one of these addresses:')
        local_addresses = ['localhost'] + get_local_addresses()
        for addr in local_addresses:
            click.echo(addr)


def select_http_port() -> int:
    """Selects port for http server."""
    port = get_def_http_port()
    res: int = click.prompt('Enter a desired HTTP port (press ENTER for default)',
                            default=str(port))
    return res


def select_projector_port() -> int:
    """Selects port for projector server."""
    port = get_def_projector_port()
    res: int = click.prompt('Enter a desired Projector port (press ENTER for default)',
                            default=str(port))
    return res


def edit_config(config: RunConfig) -> RunConfig:
    """Edits existing config."""
    prompt = 'Enter the path to IDE (press ENTER for default)'
    config.path_to_app = click.prompt(prompt, default=config.path_to_app)

    prompt = 'Enter a HTTP listening address (press ENTER for default)'
    config.http_address = click.prompt(prompt, default=config.http_address)

    prompt = 'Enter a HTTP port (press ENTER for default)'
    config.http_port = click.prompt(prompt, default=str(config.http_port))

    prompt = 'Enter a Projector port (press ENTER for default)'
    config.projector_port = click.prompt(prompt, default=str(config.projector_port))

    return config


def make_run_config(config_name: str, app_path: Optional[str] = None) -> RunConfig:
    """Creates run config with specified app_path."""
    if app_path is None:
        app_path = select_app_path()

    if app_path is None:
        print('IDE was not selected, exiting...')
        sys.exit(1)

    http_address = select_http_address('localhost')
    http_port = select_http_port()
    projector_port = select_projector_port()
    secure_config = click.prompt(
        'Use secure connection '
        '(this option requires installing a projector\'s certificate to browser)? [y/n]',
        type=bool)

    token = generate_token() if secure_config else ''
    return RunConfig(config_name, expanduser(app_path), projector_port,
                     http_address, http_port, token)


class UserInstallInput:
    """Represents user answers during install session"""

    def __init__(self, config_name: str, http_address: str, http_port: int,
                 projector_port: int, do_run: bool, secure_config: bool) -> None:
        self.config_name: str = config_name
        self.http_address: str = http_address
        self.http_port: int = http_port
        self.projector_port: int = projector_port
        self.do_run = do_run
        self.secure_config = secure_config


def get_user_install_input(config_name_hint: str, auto_run: bool) -> Optional[UserInstallInput]:
    """Interactively creates user input"""
    config_name = select_new_config_name(config_name_hint)

    if not config_name:
        return None

    http_address = select_http_address('localhost')
    http_port = select_http_port()
    projector_port = select_projector_port()
    do_run = True if auto_run else click.prompt('Would you like to run installed ide? [y/n]',
                                                type=bool)

    secure_config = click.prompt(
        'Use secure connection '
        '(this option requires installing a projector\'s certificate to browser)? [y/n]',
        type=bool)

    return UserInstallInput(config_name, http_address, http_port, projector_port,
                            do_run, secure_config)


def make_config_from_input(inp: UserInstallInput) -> RunConfig:
    """Makes run config from user input"""
    token = generate_token() if inp.secure_config else ''
    return RunConfig(inp.config_name, '', inp.projector_port,
                     inp.http_address, inp.http_port, token)



# Copyright 2000-2020 JetBrains s.r.o.
# Use of this source code is governed by the Apache 2.0 license that can be found
# in the LICENSE file.

"""Real actions performed by projector script."""

import shutil
import signal
import subprocess
import sys
from typing import Optional
from os import path, system, uname

from .apps import get_compatible_apps, get_app_path, get_installed_apps, get_product_info, \
    unpack_app
from .secure_config import get_ca_crt_file, is_secure

from .utils import download_file

from .dialogs import select_compatible_app, select_new_config_name, list_configs, \
    find_apps, edit_config, list_apps, select_installed_app, select_run_config, make_run_config, \
    get_user_install_input, make_config_from_input

from .global_config import get_http_dir, get_download_cache_dir, get_path_to_projector_log, \
    RunConfig
from .http_server_process import HttpServerProcess
from .ide_configuration import install_projector_markdown_for, forbid_updates_for
from .run_config import get_run_configs, get_run_script_path, validate_run_config, \
    save_config, delete_config, rename_config, make_config_name, get_configs_with_app, \
    update_markdown_plugin


def do_list_config(pattern: Optional[str] = None) -> None:
    """Displays existing run configs names."""
    list_configs(pattern)


def do_show_config(pattern: Optional[str] = None) -> None:
    """Shows details on run config.
    If given config name does not match unique config, runs interactive
     procedure to select it.
    """
    config_name, run_config = select_run_config(pattern)
    print(f'Configuration: {config_name}')
    print(f'IDE path: {run_config.path_to_app}')
    print(f'HTTP address: {run_config.http_address}')
    print(f'HTTP port: {run_config.http_port}')
    print(f'Projector port: {run_config.projector_port}')

    product_info = get_product_info(run_config.path_to_app)
    print(f'Product info: {product_info.name}, '
          f'version={product_info.version}, '
          f'build={product_info.build_number}')


def is_wsl() -> bool:
    """Returns True if script is run in WSL environment."""
    return uname().release.find('microsoft') != -1


def do_run_browser(url: str) -> None:
    """Starts default browser and opens provided url in WSL."""
    system(f'cmd.exe /c start {url} 2> /dev/null')


def wsl_warning() -> None:
    """Show warning for WSL environment."""
    print('It seems that you are using WSL environment.')
    print('WSL is still experimental technology, and if you experience any issues accessing '
          'projector from browser please refer to Projector README file: '
          'https://github.com/JetBrains/projector-installer#resolving-wsl-issues')


def get_access_url(run_config: RunConfig) -> str:
    """Returns access URL for given config"""
    if is_secure(run_config):
        return f'https://{run_config.http_address}:{run_config.http_port}/'

    return f'http://{run_config.http_address}:{run_config.http_port}/'


# noinspection PyShadowingNames
def do_run_config(config_name: Optional[str] = None, run_browser: bool = True) -> None:
    """Executes specified config. If given name does not specify
    config, runs interactive selection procedure."""
    config_name, run_config = select_run_config(config_name)

    print(f'Configuration name: {config_name}')
    run_script_name = get_run_script_path(config_name)

    if not path.isfile(run_script_name):
        print(f'Cannot find file {run_script_name}')
        print(f'To fix, try: projector config edit {config_name}')
        sys.exit(1)

    projector_log = open(get_path_to_projector_log(), 'a')

    projector_process = subprocess.Popen([f'{run_script_name}'],
                                         stdout=projector_log,
                                         stderr=projector_log)

    http_process = HttpServerProcess(run_config, get_http_dir())

    def signal_handler(*args):  # type: ignore
        # pylint: disable=unused-argument
        if http_process and http_process.is_alive():
            http_process.terminate()

        projector_log.close()

        print('\nExiting...')
        sys.exit(0)

    signal.signal(signal.SIGINT, signal_handler)

    http_process.start()
    access_url = get_access_url(run_config)
    print(f'HTTP process PID={http_process.pid}')
    print(f'To access your IDE, open {access_url} in your browser')

    if is_secure(run_config):
        print('If browser warns on unsecure connection, install projector certificate:')
        print(get_ca_crt_file())
        print('Refer to: ')
        print('https://github.com/JetBrains/projector-installer/blob/master/'
              'README.md#Secure-connection')

    print('Exit IDE or press Ctrl+C to stop Projector.')

    if run_browser:
        if is_wsl():
            wsl_warning()
            do_run_browser(access_url)

    projector_process.wait()

    print('Exiting...')

    if http_process and http_process.is_alive():
        http_process.terminate()

    projector_log.close()


def do_add_config(hint: Optional[str], app_path: Optional[str] = None) -> None:
    """
    Adds new run config. If auto_run = True, runs it without questions.
    Asks user otherwise.
    """
    config_name = select_new_config_name(hint)

    if config_name is None:
        print('Configuration name was not selected, exiting...')
        sys.exit(1)

    run_config = make_run_config(config_name, app_path)

    if run_config.path_to_app is None:
        print('IDE was not selected, exiting...')
        sys.exit(1)

    try:
        validate_run_config(run_config)
    except ValueError as exception:
        print(f'Wrong configuration parameters: {str(exception)}, exiting ...')
        sys.exit(1)

    save_config(run_config)


def do_remove_config(config_name: Optional[str] = None) -> None:
    """Selects (if necessary) and removes selected run config."""
    config_name, _ = select_run_config(config_name)
    print(f'Removing configuration {config_name}')
    delete_config(config_name)
    print('done.')


def do_edit_config(config_name: Optional[str] = None) -> None:
    """Selects (if necessary) and edits selected run config."""
    config_name, run_config = select_run_config(config_name)
    print(f'Edit configuration {config_name}')
    run_config = edit_config(run_config)

    try:
        validate_run_config(run_config)
    except ValueError as exception:
        print(f'Wrong configuration parameters: {str(exception)}, exiting ...')
        sys.exit(1)

    save_config(run_config)

    print('done.')


def do_rename_config(from_name: str, to_name: str) -> None:
    """Renames run config."""
    run_configs = get_run_configs()

    if from_name not in run_configs:
        print(f'Configuration name {from_name} does not exist, exiting...')
        sys.exit(1)

    if from_name == to_name:
        print('Cannot rename configuration to the same name, exiting...')
        sys.exit(1)

    if to_name in run_configs:
        print(f'Cannot rename to {to_name} - the configuration already exists, exiting...')
        sys.exit(1)

    rename_config(from_name, to_name)


def do_update_markdown_plugin(config_name: Optional[str] = None) -> None:
    """Performs markdown plugin update"""
    config_name, run_config = select_run_config(config_name)
    print(f'Updating markdown plugin in configuration {config_name}')

    update_markdown_plugin(run_config)


def do_find_app(pattern: Optional[str] = None) -> None:
    """Prints known projector-compatible apps."""
    find_apps(pattern)


def do_list_app(pattern: Optional[str] = None) -> None:
    """Prints apps installed by projector."""
    list_apps(pattern)


def do_install_app(app_name: Optional[str], auto_run: bool = False, allow_updates: bool = False,
                   run_browser: bool = True) -> None:
    """Installs specified app."""
    apps = get_compatible_apps(app_name)

    if len(apps) == 0:
        print('There are no known IDEs matched to the given name.')
        if app_name is None:
            print('Try to reinstall Projector.')
        print('Exiting...')
        sys.exit(1)

    if len(apps) > 1:
        app_name = select_compatible_app(app_name)

        if app_name is None:
            print('IDE was not selected, exiting...')
            sys.exit(1)
    else:
        app_name = apps[0].name

    apps = get_compatible_apps(app_name)
    app = apps[0]
    config_name_hint = make_config_name(app.name)
    user_input = get_user_install_input(config_name_hint, auto_run)

    if user_input is None:
        print('Config parameters was not specified, exiting ...')
        sys.exit(1)

    print(f'Installing {app.name}')

    try:
        path_to_dist = download_file(app.url, get_download_cache_dir())
    except IOError as error:
        print(f'Unable to write downloaded file, try again later: {str(error)}. Exiting ...')
        sys.exit(1)

    try:
        app_name = unpack_app(path_to_dist)
    except IOError as error:
        print(f'Unable to extract the archive: {str(error)}, exiting...')
        sys.exit(1)

    app_path = get_app_path(app_name)
    install_projector_markdown_for(app_path)

    if not allow_updates:
        forbid_updates_for(app_path)

    run_config = make_config_from_input(user_input)
    run_config.path_to_app = app_path

    try:
        validate_run_config(run_config)
    except ValueError as exception:
        print(f'Wrong configuration parameters: {str(exception)}, exiting ...')
        sys.exit(1)

    save_config(run_config)

    if user_input.do_run:
        do_run_config(run_config.name, run_browser)


def do_uninstall_app(app_name: Optional[str] = None) -> None:
    """Uninstalls specified app."""
    apps = get_installed_apps(app_name)

    if len(apps) == 0:
        print('There are no installed apps matched the given name. Exiting...')
        sys.exit(1)

    if len(apps) > 1:
        app_name = select_installed_app(app_name)
    else:
        app_name = apps[0]

    if app_name is None:
        print('IDE was not selected, exiting...')
        sys.exit(1)

    configs = get_configs_with_app(app_name)

    if configs:
        print(f'Unable to uninstall {app_name} - there are configurations with this IDE:')
        for config in configs:
            print(config)
        print('Try to remove configurations first. Exiting...')
        sys.exit(1)

    print(f'Uninstalling {app_name}')
    app_path = get_app_path(app_name)
    shutil.rmtree(app_path)
    print('done.')



# Copyright 2000-2020 JetBrains s.r.o.
# Use of this source code is governed by the Apache 2.0 license that can be found
# in the LICENSE file.


"""
Global configuration constants, variables and functions.
"""

import json
import sys
from typing import List
from shutil import rmtree
from socket import timeout
from os.path import dirname, join, expanduser, abspath

from .utils import create_dir_if_not_exist, download_file, get_file_name_from_url

USER_HOME: str = expanduser('~')
INSTALL_DIR: str = dirname(abspath(__file__))
DEF_HTTP_PORT: int = 8889
DEF_PROJECTOR_PORT: int = 9999
COMPATIBLE_IDE_FILE: str = 'compatible_ide.json'
PROJECTOR_LOG_FILE: str = 'projector.log'
DEF_CONFIG_DIR: str = '.projector'
BUNDLED_DIR: str = 'bundled'
SERVER_DIR: str = 'server'
CLIENT_DIR: str = 'client'
PLUGIN_DIR: str = 'projector-markdown-plugin'
config_dir: str = join(USER_HOME, DEF_CONFIG_DIR)


def get_path_to_license() -> str:
    """Returns full path to license file"""
    return join(INSTALL_DIR, 'LICENSE.txt')


def get_path_to_projector_log() -> str:
    """Returns full path to projector log file"""
    return join(config_dir, PROJECTOR_LOG_FILE)


def get_apps_dir() -> str:
    """Returns full path to applications directory."""
    return join(config_dir, 'apps')


def get_run_configs_dir() -> str:
    """Returns full path to run configs directory."""
    return join(config_dir, 'configs')


def get_download_cache_dir() -> str:
    """Returns full path to download cache directory."""
    return join(config_dir, 'cache')


def get_ssl_dir() -> str:
    """Returns full path to ssl directory."""
    return join(config_dir, 'ssl')


def get_compatible_ide_file() -> str:
    """Returns full path to compatible ide file."""
    return join(INSTALL_DIR, COMPATIBLE_IDE_FILE)


class CompatibleApp:
    """Compatible application entry."""

    def __init__(self, name: str, url: str) -> None:
        self.name: str = name
        self.url: str = url

    def __eq__(self, other: object) -> bool:
        if isinstance(other, CompatibleApp):
            return self.name == other.name and self.url == other.url

        return False


COMPATIBLE_APPS: List[CompatibleApp] = []


class RunConfig:
    """Run config dataclass"""

    def __init__(self, own_name: str, path_to_app: str, projector_port: int,
                 http_address: str, http_port: int, token: str) -> None:
        self.name: str = own_name
        self.path_to_app: str = path_to_app
        self.projector_port: int = projector_port
        self.http_address: str = http_address
        self.http_port: int = http_port
        self.token = token


def load_compatible_apps_from_file(file_name: str) -> None:
    """Loads compatible apps file to memory."""
    with open(file_name, 'r') as file:
        data = json.load(file)

    for entry in data:
        app = CompatibleApp(entry['name'], entry['url'])
        if app not in COMPATIBLE_APPS:
            COMPATIBLE_APPS.append(app)


COMPATIBLE_IDE_FILE_URL: str = \
    'https://raw.githubusercontent.com/JetBrains/projector-installer/master/' \
    'projector_installer/compatible_ide.json'


def download_compatible_apps() -> str:
    """Downloads compatible ide json file from github repository."""
    try:
        download_file(COMPATIBLE_IDE_FILE_URL, get_download_cache_dir(), timeout=3, silent=True)
        name = get_file_name_from_url(COMPATIBLE_IDE_FILE_URL)
        file_name = join(get_download_cache_dir(), name)

        return file_name
    except timeout:
        return ''


def load_compatible_apps() -> None:
    """Loads compatible apps dictionary from bundled file and github-stored Json"""
    file_name = get_compatible_ide_file()
    load_compatible_apps_from_file(file_name)
    github_file = download_compatible_apps()

    if github_file:
        load_compatible_apps_from_file(github_file)


def init_compatible_apps() -> None:
    """Initializes compatible apps list."""
    try:
        load_compatible_apps()
    except IOError as error:
        print(f'Cannot load compatible ide file: {str(error)}. Exiting...')
        sys.exit(2)


def get_http_dir() -> str:
    """Returns dir with client files."""
    return join(INSTALL_DIR, BUNDLED_DIR, CLIENT_DIR)


def get_projector_server_dir() -> str:
    """Returns directory with projector server jar"""
    return join(INSTALL_DIR, BUNDLED_DIR, SERVER_DIR)


def get_projector_markdown_plugin_dir() -> str:
    """Returns directory with projector markdown plugin."""
    return join(INSTALL_DIR, BUNDLED_DIR, PLUGIN_DIR)


def init_config_dir() -> None:
    """Initializes global config directory."""
    # pylint: disable=W0703
    try:
        create_dir_if_not_exist(config_dir)
        create_dir_if_not_exist(get_apps_dir())
        create_dir_if_not_exist(get_run_configs_dir())
        create_dir_if_not_exist(get_download_cache_dir())
        init_compatible_apps()
    except Exception as exception:
        print(f'Error during initialization: {str(exception)}, cleanup ...')
        rmtree(config_dir)
        sys.exit(1)



# Copyright 2000-2020 JetBrains s.r.o.
# Use of this source code is governed by the Apache 2.0 license that can be found
# in the LICENSE file.


"""Command line interface to projector-installer"""
from os import path
from os.path import expanduser, realpath, expandvars
from typing import Any, Optional
import click

from . import global_config
from .global_config import init_config_dir, init_compatible_apps

from .actions import do_install_app, do_uninstall_app, do_find_app, do_list_app, do_run_config, \
    do_list_config, do_show_config, do_add_config, do_remove_config, do_edit_config, \
    do_rename_config, do_update_markdown_plugin
from .license import display_license


def is_first_start() -> bool:
    """Detects first app start."""
    return not path.isdir(global_config.config_dir)


@click.group(invoke_without_command=True)
@click.pass_context
@click.version_option()
@click.option('--config-directory', type=click.Path(),
              default=global_config.config_dir,
              help='Path to configuration directory')
def projector(ctx: Any, config_directory: str) -> None:
    """
    This script helps to install, manage, and run JetBrains IDEs with Projector.
    """

    global_config.config_dir = realpath(expandvars(expanduser(config_directory)))

    if is_first_start():
        display_license()
        init_config_dir()
        print('Please select IDE to install:')
        do_install_app(None, auto_run=True, allow_updates=False, run_browser=True)
    elif not ctx.invoked_subcommand:
        click.echo(ctx.get_help())
    else:
        init_compatible_apps()


@projector.group()
def ide() -> None:
    """
    JetBrains IDEs management commands
    """


@click.command(short_help='Find Projector-compatible IDE')
@click.argument('pattern', type=click.STRING, required=False)
def find_app(pattern: Optional[str]) -> None:
    """projector ide find [pattern]

    Find projector-compatible IDE with the name matching to the given pattern.
    If no pattern is specified, finds all the compatible IDEs.
    """
    do_find_app(pattern)


ide.add_command(find_app, name='find')


@click.command(short_help='Install and configure selected IDE')
@click.argument('ide_name', type=click.STRING, required=False)
@click.option('--auto-run', default=False, is_flag=True,
              help='Run installed IDE without confirmation.')
@click.option('--allow-updates', default=False, is_flag=True,
              help='Allow updates of installed IDE.')
@click.option('--run-browser/--no-browser', default=True,
              help='Auto run browser in WSL environment.')
def install_app(ide_name: Optional[str], auto_run: bool, allow_updates: bool,
                run_browser: bool) -> None:
    """projector ide install [ide_name]

    Parameter ide_name is the name of IDE to install.
    If no IDE name is given or the pattern is ambiguous, guides the user through the
    install process.
    """
    do_install_app(ide_name, auto_run, allow_updates, run_browser)


ide.add_command(install_app, name='install')


@ide.command(short_help='Uninstall selected IDE')
@click.argument('name_pattern', type=click.STRING, required=False)
def uninstall(name_pattern: Optional[str]) -> None:
    """projector ide install [ide_name_pattern]

    Parameter ide_name_pattern is matched to the name of IDE to uninstall.
    If no name pattern is given or the pattern is ambiguous, guides the user through the
    uninstall process.
    """
    do_uninstall_app(name_pattern)


@click.command(short_help='Display installed IDEs')
@click.argument('pattern', type=click.STRING, required=False)
def list_apps(pattern: Optional[str]) -> None:
    """projector ide list [pattern]

    Displays installed IDEs whose names matches to given pattern.
    If no pattern is given, lists all installed IDEs.
    """
    do_list_app(pattern)


ide.add_command(list_apps, name='list')


@projector.group()
def config() -> None:
    """
    Configuration management commands
    """


@click.command(short_help='Run selected config')
@click.argument('config_name', type=click.STRING, required=False)
@click.option('--run-browser/--no-browser', default=True,
              help='Auto run browser in WSL environment.')
def run_config(config_name: Optional[str], run_browser: bool) -> None:
    """projector config run config_name_pattern

    Parameter config_name_pattern specifies the configuration to run.
    If no configuration specified or the pattern is ambiguous, selects a configuration
    interactively.
    """
    do_run_config(config_name, run_browser)


config.add_command(run_config, name='run')


@click.command(short_help='List configurations')
@click.argument('pattern', required=False)
def list_config(pattern: Optional[str]) -> None:
    """projector config list [pattern]

    Displays configurations whose names matches to the given pattern.
    If no pattern is given, lists all the configurations.
    """
    do_list_config(pattern)


config.add_command(list_config, name='list')


@config.command(short_help='Show selected configuration details')
@click.argument('config_name', type=click.STRING, required=False)
def show(config_name: Optional[str]) -> None:
    """projector config show [config_name]

    Parameter config_name specifies a desired configuration.
    If not given or ambiguous, selects a configuration interactively.
    """
    do_show_config(config_name)


@config.command(short_help='Add new configuration')
@click.argument('config_name', type=click.STRING, required=False)
@click.argument('ide_path', type=click.STRING, required=False)
def add(config_name: Optional[str], ide_path: Optional[str]) -> None:
    """projector config add [config_name]

    Add a new configuration.
    """
    do_add_config(config_name, ide_path)


@config.command(short_help='Remove configuration')
@click.argument('config_name', type=click.STRING, required=False)
def remove(config_name: Optional[str]) -> None:
    """projector config remove [config_name]

    Remove an existing configuration.
    """
    do_remove_config(config_name)


@config.command(short_help='Change existing configuration')
@click.argument('config_name', type=click.STRING, required=False)
def edit(config_name: Optional[str]) -> None:
    """projector config edit [config_name]

    Change an existing configuration.
    """
    do_edit_config(config_name)


@config.command(short_help='Rename existing configuration')
@click.argument('from_name', type=click.STRING, required=True)
@click.argument('to_name', type=click.STRING, required=True)
def rename(from_name: str, to_name: str) -> None:
    """projector config rename from_config_name to_config_name

    Rename an existing configuration.
    """
    do_rename_config(from_name, to_name)


@config.command(short_help='Update markdown plugin in existing configuration')
@click.argument('config_name', type=click.STRING, required=False)
def update_markdown_plugin(config_name: Optional[str]) -> None:
    """projector config update-markdown-plugin [config_name]

    Update markdown plugin in existing configuration.
    """
    do_update_markdown_plugin(config_name)


config.add_command(update_markdown_plugin, name='update-markdown-plugin')


# Projector commands shortcuts

@projector.command(short_help='Run selected configuration')
@click.argument('config_name', type=click.STRING, required=False)
@click.option('--run-browser/--no-browser', default=True,
              help='Auto run browser in WSL environment.')
def run(config_name: Optional[str], run_browser: bool) -> None:
    """projector run config_name

    Shortcut for projector config run config_name
    """
    do_run_config(config_name, run_browser)


@projector.command(short_help='Install and configure selected IDE')
@click.argument('ide_name', type=click.STRING, required=False)
@click.option('--auto-run', default=False, is_flag=True,
              help='Run installed ide without confirmation.')
@click.option('--allow-updates', default=False, is_flag=True,
              help='Allow updates of installed IDE.')
@click.option('--run-browser/--no-browser', default=True,
              help='Auto run browser in WSL environment.')
def install(ide_name: Optional[str], auto_run: bool, allow_updates: bool, run_browser: bool) \
        -> None:
    """projector install [ide_name]

    Shortcut for projector ide install [ide_name]
    """
    do_install_app(ide_name, auto_run, allow_updates, run_browser)


@projector.command(short_help='Find available IDEs')
@click.argument('pattern', type=click.STRING, required=False)
def find(pattern: Optional[str]) -> None:
    """projector find [pattern]

    Shortcut for projector ide find [pattern]
    """
    do_find_app(pattern)



# Copyright 2000-2020 JetBrains s.r.o.
# Use of this source code is governed by the Apache 2.0 license that can be found
# in the LICENSE file.


"""projector-installer version"""

__version__: str = '0.1.8'



# Copyright 2000-2020 JetBrains s.r.o.
# Use of this source code is governed by the Apache 2.0 license that can be found
# in the LICENSE file.


"""
Misc utility functions.
"""

import tarfile
import zipfile
from os import listdir, remove, makedirs
from os.path import join, isfile, getsize, basename, isdir
from shutil import copy
from urllib.request import urlopen
from typing import Optional

from click import progressbar, echo

CHUNK_SIZE = 4 * 1024 * 1024
PROGRESS_BAR_WIDTH = 50
PROGRESS_BAR_TEMPLATE = '[%(bar)s]  %(info)s'


def create_dir_if_not_exist(dir_name: str) -> None:
    """Creates given directory with all parents if it is not exist."""
    if not isdir(dir_name):
        makedirs(dir_name, mode=0o700, exist_ok=True)


def remove_file_if_exist(file_name: str) -> None:
    """Removes existing file"""
    if isfile(file_name):
        remove(file_name)


def copy_all_files(source: str, destination: str) -> None:
    """Copies all files from source directory to destination."""
    for file_name in listdir(source):
        from_path = join(source, file_name)
        to_path = join(destination, file_name)

        if isfile(from_path):
            copy(from_path, to_path)


def get_file_name_from_url(url: str) -> str:
    """
    Extracts file name from URL.
    """
    parts = url.split('/')
    result = parts[-1]
    pos = result.find('?')

    if pos != -1:
        result = result[:pos]

    return result


def download_file(url: str, destination: str, timeout: Optional[int] = None,
                  silent: Optional[bool] = False) -> str:
    """
    Downloads file by given URL to destination dir.
    """
    file_name = get_file_name_from_url(url)
    file_path = join(destination, file_name)

    with urlopen(url, timeout=timeout) as resp:
        code: int = resp.getcode()

        if code != 200:
            raise IOError(f'Bad HTTP response code: {code}')

        total = int(resp.getheader('Content-Length'))

        if not isfile(file_path) or getsize(file_path) != total:

            if not silent:
                echo(f'Downloading {file_name}')

            with open(file_path, 'wb') as file, \
                    progressbar(length=total,
                                width=PROGRESS_BAR_WIDTH,
                                bar_template=PROGRESS_BAR_TEMPLATE) as progress_bar:

                while True:
                    chunk = resp.read(CHUNK_SIZE)

                    if not chunk:
                        break

                    file.write(chunk)

                    if not silent:
                        progress_bar.update(len(chunk))

    return file_path


def unpack_tar_file(file_path: str, destination: str) -> str:
    """ Unpacks given file in destination directory. """
    print(f'Unpacking {basename(file_path)}')

    with tarfile.open(file_path) as tar_file:
        members = tar_file.getmembers()
        dir_name = members[0].name.split('/')[0]

        with progressbar(length=len(members), width=PROGRESS_BAR_WIDTH,
                         bar_template=PROGRESS_BAR_TEMPLATE) as progress_bar:
            for member in members:
                tar_file.extract(member, destination)
                progress_bar.update(1)

    return dir_name


def unpack_zip_file(file_path: str, destination: str) -> str:
    """ Unpacks given file in destination directory. """
    print(f'Unpacking {basename(file_path)}')

    with zipfile.ZipFile(file_path) as zip_file:
        file_names = zip_file.namelist()
        dir_name = file_names[0].split('/')[0]

        with progressbar(length=len(file_names), width=PROGRESS_BAR_WIDTH,
                         bar_template=PROGRESS_BAR_TEMPLATE) as progress_bar:
            for file_name in file_names:
                zip_info = zip_file.getinfo(file_name)
                zip_file.extract(zip_info, destination)
                progress_bar.update(1)

    return dir_name



# Copyright 2000-2020 JetBrains s.r.o.
# Use of this source code is governed by the Apache 2.0 license that can be found
# in the LICENSE file.


"""HTTP server process module."""

from os import chdir
from typing import Any, Optional
from http.server import SimpleHTTPRequestHandler
from multiprocessing import Process

import socketserver
import socket
import ssl

from .global_config import RunConfig
from .secure_config import get_http_crt_file, get_http_key_file, is_secure


class NoLogServer(SimpleHTTPRequestHandler):
    """Custom http server class to serve static projector client files."""
    address: str = ''
    port: str = ''
    projector_port: str = ''
    token: str = ''

    def log_message(self, *args: Any) -> None:
        pass

    def do_GET(self) -> None:
        try:
            if self.need_redirect():
                self.send_response(301)
                self.send_header('Location', NoLogServer.redirect_url())
                self.end_headers()
            else:
                super().do_GET()
        except socket.error:
            pass

    @classmethod
    def init_with(cls, run_config: RunConfig) -> None:
        """Initializes server with run config params"""
        cls.address = run_config.http_address
        cls.port = str(run_config.http_port)
        cls.projector_port = str(run_config.projector_port)
        cls.token = run_config.token

    @classmethod
    def redirect_url(cls) -> str:
        """Constructs redirect url."""

        if len(cls.token) > 0:  # secure connection
            return f'https://{cls.address}:{cls.port}/index.html?' \
                   f'host={cls.address}&port={cls.projector_port}'

        return f'http://{cls.address}:{cls.port}/index.html?' \
               f'host={cls.address}&port={cls.projector_port}'

    def is_empty_path(self) -> bool:
        """Checks if current path is empty."""
        return not self.path or self.path == '/'

    def path_contains_index_html(self) -> bool:
        """Checks if path contains index.html"""
        return self.path.find('index.html') != -1

    def path_contains_projector_port(self) -> bool:
        """Checks if path contains projector port."""
        return self.path.find(f'port={NoLogServer.projector_port}') != -1

    def need_redirect(self) -> bool:
        """Checks if we need redirect request."""
        if self.is_empty_path():
            return True

        if not self.path_contains_index_html():
            return False

        return not self.path_contains_projector_port()


class HttpServerProcess(Process):
    """Http server process class."""

    def __init__(self, run_config: RunConfig, directory: str) -> None:
        super().__init__(daemon=True)
        self.run_config = run_config
        self.directory = directory
        self.httpd: Optional[socketserver.BaseServer] = None

    def run(self) -> None:
        chdir(self.directory)
        NoLogServer.init_with(self.run_config)
        socketserver.TCPServer.allow_reuse_address = True

        with socketserver.TCPServer((self.run_config.http_address,
                                     self.run_config.http_port),
                                    NoLogServer) as httpd:
            if is_secure(self.run_config):
                context = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)
                context.load_cert_chain(get_http_crt_file(self.run_config.name),
                                        get_http_key_file(self.run_config.name))
                httpd.socket = context.wrap_socket(httpd.socket, server_side=True)

            self.httpd = httpd

            try:
                httpd.serve_forever()
            except socket.error:
                httpd.shutdown()
                httpd.server_close()
                self.httpd = None

    def terminate(self) -> None:
        if self.httpd:
            self.httpd.shutdown()
            self.httpd.server_close()

        super().terminate()



# Copyright 2000-2020 JetBrains s.r.o.
# Use of this source code is governed by the Apache 2.0 license that can be found
# in the LICENSE file.


"""
Ide configuration routines.
"""

from os.path import join, isfile, basename, dirname
from distutils.dir_util import copy_tree
from .global_config import get_projector_markdown_plugin_dir
from .apps import get_config_dir, get_plugin_dir, get_bin_dir
from .utils import create_dir_if_not_exist


def is_disabled(file_name: str, plugin_name: str) -> bool:
    """Checks if given plugin is already disabled"""
    if not isfile(file_name):
        return False

    with open(file_name, 'r') as file:
        lines = [line.strip() for line in file]
        return plugin_name in lines


def disable_plugin(file_name: str, plugin_name: str) -> None:
    """Disables specified plugin"""
    directory = dirname(file_name)
    create_dir_if_not_exist(directory)

    with open(file_name, 'a') as file:
        file.write(f'{plugin_name}')


DISABLED_PLUGINS_FILE = 'disabled_plugins.txt'
MARKDOWN_PLUGIN_NAME = 'org.intellij.plugins.markdown'


def disable_markdown_plugin(app_path: str) -> None:
    """Disables markdown plugin"""
    config_dir = get_config_dir(app_path)
    file_name = join(config_dir, DISABLED_PLUGINS_FILE)

    if not is_disabled(file_name, MARKDOWN_PLUGIN_NAME):
        disable_plugin(file_name, MARKDOWN_PLUGIN_NAME)


def install_own_markdown_plugin(app_path: str) -> None:
    """Install projector markdown plugin"""
    destination_dir = get_plugin_dir(app_path)
    destination_dir = join(destination_dir, basename(get_projector_markdown_plugin_dir()))

    copy_tree(get_projector_markdown_plugin_dir(), destination_dir)


def install_projector_markdown_for(app_path: str) -> None:
    """Install projector markdown plugin for specified application."""
    disable_markdown_plugin(app_path)
    install_own_markdown_plugin(app_path)


IDEA_PROPERTIES_FILE = 'idea.properties'
FORBID_UPDATE_STRING = 'ide.no.platform.update=Projector'


def forbid_updates_for(app_path: str) -> None:
    """Forbids IDEA platform update for specified app."""
    bin_dir = get_bin_dir(app_path)
    prop_file = join(bin_dir, IDEA_PROPERTIES_FILE)

    with open(prop_file, 'a') as file:
        file.write(f'{FORBID_UPDATE_STRING}')



# Copyright 2000-2020 JetBrains s.r.o.
# Use of this source code is governed by the Apache 2.0 license that can be found
# in the LICENSE file.


"""projector-installer setup file."""

from shutil import copyfile, rmtree
from os.path import isfile, join, isdir
from os import remove
from typing import List
from setuptools import setup  # type: ignore
from setuptools.command.install import install  # type: ignore
from setuptools import Command

from projector_installer.utils import create_dir_if_not_exist, download_file, unpack_zip_file, \
    get_file_name_from_url, copy_all_files

from projector_installer.global_config import BUNDLED_DIR, SERVER_DIR, CLIENT_DIR


def copy_license() -> None:
    """Copy license file to package"""
    if isfile('license/LICENSE.txt'):
        copyfile('license/LICENSE.txt', 'projector_installer/LICENSE.txt')


with open('requirements.txt') as f:
    requirements = f.read().splitlines()

PACKAGE_DIR = 'projector_installer'
bundled_dir = join(PACKAGE_DIR, BUNDLED_DIR)
server_dir = join(bundled_dir, SERVER_DIR)
client_dir = join(bundled_dir, CLIENT_DIR)
plugin_dir = bundled_dir

PROJECTOR_SERVER_URL: str = 'https://github.com/JetBrains/projector-server/releases/' \
                            'download/v0.45.9/projector-server-v0.45.9.zip'

PROJECTOR_CLIENT_URL: str = 'https://github.com/JetBrains/projector-client/releases/' \
                            'download/v0.46.10/projector-client-web-distribution-v0.46.10.zip'

MARKDOWN_PLUGIN_URL: str = 'https://github.com/JetBrains/projector-markdown-plugin/releases/' \
                           'download/v0.42.2/projector-markdown-plugin-v0.42.2.zip'


def download_client(to_directory: str) -> None:
    """Download and  unpack projector client"""
    download_file(PROJECTOR_CLIENT_URL, to_directory)
    file_path = join(to_directory, get_file_name_from_url(PROJECTOR_CLIENT_URL))
    unpack_zip_file(file_path, to_directory)
    remove(file_path)


def download_server(to_dir: str) -> None:
    """Download and  unpack projector server"""
    download_file(PROJECTOR_SERVER_URL, to_dir)
    file_path = join(to_dir, get_file_name_from_url(PROJECTOR_SERVER_URL))
    dir_name = unpack_zip_file(file_path, to_dir)
    temp_dir = join(to_dir, dir_name)
    jars_path = join(temp_dir, 'lib')
    copy_all_files(jars_path, to_dir)
    rmtree(temp_dir)
    remove(file_path)


def download_plugin(to_dir: str) -> None:
    """Downloads and installs projector markdown plugin"""
    download_file(MARKDOWN_PLUGIN_URL, to_dir)
    file_path = join(to_dir, get_file_name_from_url(MARKDOWN_PLUGIN_URL))
    unpack_zip_file(file_path, to_dir)
    remove(file_path)


def download_bundled_data() -> None:
    """Downloads data to bundle in package"""

    if not isdir(bundled_dir):
        create_dir_if_not_exist(bundled_dir)
        create_dir_if_not_exist(server_dir)
        create_dir_if_not_exist(client_dir)
        create_dir_if_not_exist(plugin_dir)
        download_client(client_dir)
        download_server(server_dir)
        download_plugin(plugin_dir)


class BundleCommand(Command):
    """Download bundled data."""
    description = 'Download bundled data.'
    user_options: List[str] = []

    def initialize_options(self) -> None:
        """Abstract method stub"""

    def finalize_options(self) -> None:
        """Abstract method stub"""

    def run(self) -> None:
        """Run command."""
        copy_license()
        download_bundled_data()


class CustomInstallCommand(install):
    """
    Customized install command.
    """

    def run(self) -> None:
        """
        Copies add data to package.
        """
        copy_license()
        download_bundled_data()
        install.run(self)


setup(
    install_requires=requirements,
    setup_requires=['click'],
    cmdclass={
        'bundle': BundleCommand,
        'install': CustomInstallCommand
    }
)



#  Copyright 2000-2020 JetBrains s.r.o.
#  Use of this source code is governed by the Apache 2.0 license that can be found
#  in the LICENSE file.

"""Secure config related stuff"""
import socket
from os.path import join, isfile
from typing import List

import re
import subprocess
import secrets
import string

from .global_config import get_ssl_dir, RunConfig, get_run_configs_dir
from .utils import create_dir_if_not_exist, remove_file_if_exist

SSL_ENV_NAME = 'ORG_JETBRAINS_PROJECTOR_SERVER_SSL_PROPERTIES_PATH'
TOKEN_ENV_NAME = 'ORG_JETBRAINS_PROJECTOR_SERVER_HANDSHAKE_TOKEN'

SSL_PROPERTIES_FILE = 'ssl.properties'
PROJECTOR_JKS_NAME = 'projector'
HTTP_SERVER = 'http_server'

DEF_TOKEN_LEN = 20
CA_NAME = 'ca'
CA_PASSWORD = '85TibAyPS3NZX3'


def generate_token(length: int = DEF_TOKEN_LEN) -> str:
    """Generates token to access server's secrets"""
    alphabet = string.ascii_letters + string.digits
    return ''.join(secrets.choice(alphabet) for i in range(length))


def get_http_crt_file(config_name: str) -> str:
    """Returns full path to http server certificate file"""
    return join(get_run_configs_dir(), config_name, f'{HTTP_SERVER}.crt')


def get_http_key_file(config_name: str) -> str:
    """Returns full path to http server key file"""
    return join(get_run_configs_dir(), config_name, f'{HTTP_SERVER}.key')


def get_http_csr_file(config_name: str) -> str:
    """Returns full path to projector server crt file"""
    return join(get_run_configs_dir(), config_name, f'{HTTP_SERVER}.csr')


def get_http_jks_file(config_name: str) -> str:
    """Returns full path to projector server crt file"""
    return join(get_run_configs_dir(), config_name, f'{HTTP_SERVER}.jks')


def get_ssl_properties_file(config_name: str) -> str:
    """Returns full path to ssl.properties file"""
    return join(get_run_configs_dir(), config_name, SSL_PROPERTIES_FILE)


def get_projector_jks_file(config_name: str) -> str:
    """Returns full path to projector server key file"""
    return join(get_run_configs_dir(), config_name, f'{PROJECTOR_JKS_NAME}.jks')


def get_projector_pkcs12_file(config_name: str) -> str:
    """Returns full path to projector server key file"""
    return join(get_run_configs_dir(), config_name, f'{PROJECTOR_JKS_NAME}.p12')


def get_projector_csr_file(config_name: str) -> str:
    """Returns full path to projector server crt file"""
    return join(get_run_configs_dir(), config_name, f'{PROJECTOR_JKS_NAME}.csr')


def get_projector_crt_file(config_name: str) -> str:
    """Returns full path to projector server crt file"""
    return join(get_run_configs_dir(), config_name, f'{PROJECTOR_JKS_NAME}.crt')


def get_ca_crt_file() -> str:
    """Returns full path to ca certificate file"""
    return join(get_ssl_dir(), f'{CA_NAME}.crt')


def get_ca_key_file() -> str:
    """Returns full path to ca key file"""
    return join(get_ssl_dir(), f'{CA_NAME}.key')


def get_ca_jks_file() -> str:
    """Returns full path to ca keystore"""
    return join(get_ssl_dir(), f'{CA_NAME}.jks')


def get_ca_pkcs12_file() -> str:
    """Returns full path to ca keystore"""
    return join(get_ssl_dir(), f'{CA_NAME}.p12')


def is_ca_exist() -> bool:
    """Checks if ca already generated"""
    ret = isfile(get_ca_jks_file())
    ret = ret and isfile(get_ca_crt_file())
    return ret


def get_ca_dist_name() -> str:
    """Returns CA Dist name"""

    return f'CN=PROJECTOR-{socket.gethostname()}-{generate_token(5)}-CA, ' \
           f'OU=Development, O=Projector, L=SPB, S=SPB, C=RU'


def get_generate_ca_command() -> List[str]:
    """Returns list of args for generate ca"""
    return ['-genkeypair', '-alias', CA_NAME,
            '-dname', get_ca_dist_name(), '-keystore', get_ca_jks_file(),
            '-keypass', CA_PASSWORD, '-storepass', CA_PASSWORD,
            '-keyalg', 'RSA', '-keysize', '4096',
            '-ext', 'KeyUsage:critical=keyCertSign',
            '-ext', 'BasicConstraints:critical=ca:true',
            '-validity', '9999'
            ]


def get_export_ca_command() -> List[str]:
    """Returns list of args for export ca.crt"""
    return ['-export', '-alias', CA_NAME, '-file', get_ca_crt_file(), '-keypass', CA_PASSWORD,
            '-storepass', CA_PASSWORD, '-keystore', get_ca_jks_file(), '-rfc']


def get_convert_ca_to_pkcs12() -> List[str]:
    """Returns list of args for convert ca to pkcs12 format"""
    return [
        '-importkeystore', '-srckeystore', get_ca_jks_file(),
        '-srcstoretype', 'JKS',
        '-srcstorepass', CA_PASSWORD,
        '-destkeystore', get_ca_pkcs12_file(),
        '-deststoretype', 'pkcs12',
        '-deststorepass', CA_PASSWORD
    ]


def get_extract_ca_key_args() -> List[str]:
    """Returns list of openssl args for extract ca key"""
    return [
        'pkcs12',
        '-in', get_ca_pkcs12_file(),
        '-nodes',
        '-nocerts',
        '-out', get_ca_key_file(),
        '-passin', f'pass:{CA_PASSWORD}'
    ]


def generate_ca(keytool_path: str) -> None:
    """Creates CA"""
    create_dir_if_not_exist(get_ssl_dir())

    cmd = [keytool_path] + get_generate_ca_command()
    subprocess.check_call(cmd, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)

    cmd = [keytool_path] + get_export_ca_command()
    subprocess.check_call(cmd, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)

    cmd = [keytool_path] + get_convert_ca_to_pkcs12()
    subprocess.check_call(cmd, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)

    cmd = ['openssl'] + get_extract_ca_key_args()
    subprocess.check_call(cmd, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)


DIST_PROJECTOR_NAME = 'CN=Idea, OU=Development, O=Idea, L=SPB, S=SPB, C=RU'


def get_projector_gen_jks_args(run_config: RunConfig) -> List[str]:
    """keytool args for projector jks generation"""
    return [
        '-genkeypair', '-alias', PROJECTOR_JKS_NAME, '-dname', DIST_PROJECTOR_NAME,
        '-keystore', get_projector_jks_file(run_config.name),
        '-keypass', run_config.token, '-storepass', run_config.token,
        '-keyalg', 'RSA', '-keysize', '4096', '-validity', '4500'
    ]


def get_projector_cert_sign_request_args(run_config: RunConfig) -> List[str]:
    """Returns list of args for request cert sign"""
    return [
        '-certreq', '-alias', PROJECTOR_JKS_NAME, '-keypass', run_config.token,
        '-storepass', run_config.token,
        '-keystore', get_projector_jks_file(run_config.name),
        '-file', get_projector_csr_file(run_config.name)
    ]


def is_ip_address(address: str) -> bool:
    """Detects if given string is IP address"""
    return re.match(
        r'^(([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\.)'
        '{3}([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])$',
        address) is not None


def get_projector_san(address: str) -> str:
    """Returns san"""
    if is_ip_address(address):
        return "IP:" + address

    return "DNS:" + address


def get_projector_cert_sign_args(run_config: RunConfig) -> List[str]:
    """Returns list of args to sign projector server cert"""
    return [
        '-gencert',
        '-alias', CA_NAME,
        '-keypass', run_config.token,
        '-storepass', CA_PASSWORD,
        '-keystore', get_ca_jks_file(),
        '-infile', get_projector_csr_file(run_config.name),
        '-outfile', get_projector_crt_file(run_config.name),
        '-ext', 'KeyUsage:critical=digitalSignature,keyEncipherment',
        '-ext', 'EKU=serverAuth',
        '-ext', f'SAN={get_projector_san(run_config.http_address)}',
        '-rfc'
    ]


def get_projector_import_ca_args(run_config: RunConfig) -> List[str]:
    """Returns list of args to import ca to projector jks"""
    return [
        '-import', '-alias', CA_NAME,
        '-file', get_ca_crt_file(),
        '-keystore', get_projector_jks_file(run_config.name),
        '-storetype', 'JKS',
        '-storepass', run_config.token,
        '-noprompt'
    ]


def get_projector_import_cert_args(run_config: RunConfig) -> List[str]:
    """Returns list of args tyo import projector cert to jks"""
    return [
        '-import', '-alias', PROJECTOR_JKS_NAME,
        '-file', get_projector_crt_file(run_config.name),
        '-keystore', get_projector_jks_file(run_config.name),
        '-storetype', 'JKS',
        '-storepass', run_config.token
    ]


def generate_projector_jks(run_config: RunConfig) -> None:
    """Generates projector jks for given config"""

    keytool_path = get_jbr_keytool(run_config.path_to_app)

    cmd = [keytool_path] + get_projector_gen_jks_args(run_config)
    subprocess.check_call(cmd, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)

    cmd = [keytool_path] + get_projector_cert_sign_request_args(run_config)
    subprocess.check_call(cmd, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)

    cmd = [keytool_path] + get_projector_cert_sign_args(run_config)
    subprocess.check_call(cmd, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)

    cmd = [keytool_path] + get_projector_import_ca_args(run_config)
    subprocess.check_call(cmd, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)

    cmd = [keytool_path] + get_projector_import_cert_args(run_config)
    subprocess.check_call(cmd, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)


DIST_HTTP_NAME = 'CN=Http, OU=Development, O=Idea, L=SPB, S=SPB, C=RU'


def get_http_gen_args(run_config: RunConfig) -> List[str]:
    """keytool args for http keypair generation"""
    return [
        '-genkeypair', '-alias', HTTP_SERVER, '-dname', DIST_HTTP_NAME,
        '-keystore', get_http_jks_file(run_config.name),
        '-keypass', run_config.token, '-storepass', run_config.token,
        '-keyalg', 'RSA', '-keysize', '4096', '-validity', '4500'
    ]


def get_openssl_generate_key_args(run_config: RunConfig) -> List[str]:
    """Get openssl generate keys arg"""
    return ['genrsa',
            '-out', get_http_key_file(run_config.name),
            '2048'
            ]


def get_openssl_subj(http_address: str) -> str:
    """Returns subject string for http server cert request"""
    return "/C=RU/ST=SPB/L=SPB/O=Projector/OU=projector-installer/CN=" + http_address


def get_openssl_generate_cert_req(run_config: RunConfig) -> List[str]:
    """Returns openssl args to genetrate cert sign request"""
    return ['req', '-new',
            '-key', get_http_key_file(run_config.name),
            '-out', get_http_csr_file(run_config.name),
            '-subj', get_openssl_subj(run_config.http_address)
            ]


def get_openssl_sign_args(run_config: RunConfig) -> List[str]:
    """Returns openssl list of args for sign cert"""
    return [
        'x509', '-req',
        '-in', get_http_csr_file(run_config.name),
        '-CA', get_ca_crt_file(),
        '-CAkey', get_ca_key_file(),
        '-CAcreateserial',
        '-out', get_http_crt_file(run_config.name),
        '-days', '4500'
    ]


def generate_http_cert(run_config: RunConfig) -> None:
    """Generates http certificate and key files"""
    cmd = ['openssl'] + get_openssl_generate_key_args(run_config)
    subprocess.check_call(cmd, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)

    cmd = ['openssl'] + get_openssl_generate_cert_req(run_config)
    subprocess.check_call(cmd, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)

    cmd = ['openssl'] + get_openssl_sign_args(run_config)
    subprocess.check_call(cmd, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)


def generate_ssl_properties_file(config_name: str, token: str) -> None:
    """Generates ssl.properties file for given config"""
    with open(get_ssl_properties_file(config_name), "w") as file:
        print('STORE_TYPE=JKS', file=file)
        print(f'FILE_PATH={get_projector_jks_file(config_name)}', file=file)
        print(f'STORE_PASSWORD={token}', file=file)
        print(f'KEY_PASSWORD={token}', file=file)


def get_jbr_keytool(path_to_app: str) -> str:
    """Returns full path to keytool for given config"""
    return join(path_to_app, 'jbr', 'bin', 'keytool')


def are_exist_server_secrets(config_name: str) -> bool:
    """Returns True if all secure connection related server config files exist"""
    ret: bool = isfile(get_projector_jks_file(config_name))
    ret = ret and isfile(get_projector_csr_file(config_name))
    ret = ret and isfile(get_projector_crt_file(config_name))
    ret = ret and isfile(get_ssl_properties_file(config_name))
    ret = ret and isfile(get_http_csr_file(config_name))
    ret = ret and isfile(get_http_crt_file(config_name))
    ret = ret and isfile(get_http_key_file(config_name))

    return ret


def remove_server_secrets(config_name: str) -> None:
    """Removes existing server secret files"""
    remove_file_if_exist(get_projector_jks_file(config_name))
    remove_file_if_exist(get_projector_csr_file(config_name))
    remove_file_if_exist(get_projector_crt_file(config_name))
    remove_file_if_exist(get_ssl_properties_file(config_name))
    remove_file_if_exist(get_http_csr_file(config_name))
    remove_file_if_exist(get_http_crt_file(config_name))
    remove_file_if_exist(get_http_key_file(config_name))


def generate_server_secrets(run_config: RunConfig) -> None:
    """Generate all secret connection related stuff for given config"""
    keytool_path = get_jbr_keytool(run_config.path_to_app)
    if not is_ca_exist():
        generate_ca(keytool_path)

    if not are_exist_server_secrets(run_config.name):  # if not all files exist
        remove_server_secrets(run_config.name)  # remove existing files
        generate_projector_jks(run_config)
        generate_ssl_properties_file(run_config.name, run_config.token)
        generate_http_cert(run_config)


def is_secure(run_config: RunConfig) -> bool:
    """Checks if secure configuration"""
    return run_config.token != ''



# Copyright 2000-2020 JetBrains s.r.o.
# Use of this source code is governed by the Apache 2.0 license that can be found
# in the LICENSE file.


"""License related functions."""

import sys
import subprocess
import click

from .global_config import get_path_to_license


def display_license() -> None:
    """
    Displays the license and asks the user to accept it.
    Exits the program if the license is not accepted.
    """
    license_file = get_path_to_license()

    subprocess.run(
        ['less', '-P Please read the license. Up/Down/Space - scroll, q - finish reading. ',
         license_file], check=False)

    accept = click.prompt('This software includes components licensed under GPLv2. Do you accept '
                          'this license? [y/n]', type=bool)

    if not accept:
        click.echo('The license was not accepted, exiting...')
        sys.exit(2)
